/**
 * Copyright 2019-2021 Britanicus <marcusbritanicus@gmail.com>
 *
 * This file is a part of DesQ project (https://gitlab.com/desq/)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * at your option, any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301, USA.
 **/

#include "Manager.hpp"

#include <desq/Xdg.hpp>
#include <desq/Utils.hpp>
#include <wlrootsqt/Registry.hpp>
#include <wlrootsqt/Application.hpp>
#include <wlrootsqt/DataControl.hpp>


DesQ::Clipboard::Manager::Manager() : QObject() {
    /** Registry */
    Wayland::Registry *registry = qApp->waylandRegistry();

    /** DataControlManager */
    mDataMgr = registry->dataControlManager();

    /** DataControlDevice and its connections */
    mDataDev = mDataMgr->getDataDevice( registry->waylandSeat() );
    /** Handle the Clipboard offer */
    connect( mDataDev, &WlrootsQt::DataControlDevice::selectionOffered,        this, &DesQ::Clipboard::Manager::handleClipboardOffer );
    /** Handle the Primary Selection offer */
    connect( mDataDev, &WlrootsQt::DataControlDevice::primarySelectionOffered, this, &DesQ::Clipboard::Manager::handleSelectionOffer );
    /** Finish the mDataDev setup */
    mDataDev->setup();

    /* Clipboard History */
    clipboard = new DesQ::Clipboard::Table( DesQ::Clipboard::Clipboard );

    /* Selection History */
    selection = new DesQ::Clipboard::Table( DesQ::Clipboard::Selection );
}


void DesQ::Clipboard::Manager::requestOffer( DesQ::Clipboard::Type type, int idx ) {
    /**
     * Even though we can deny a request, at the moment, we don't have any reason to.
     * We honour all the requests for now.
     */

    qDebug() << "Requested" << (type == DesQ::Clipboard::Clipboard ? "Clipboard" : "Selection") << "at index" << idx;

    /**
     * Special request: Set the clipboard/selection to null.
     */
    if ( idx == -1 ) {
        if ( type == DesQ::Clipboard::Clipboard ) {
            mDataDev->setSelection( nullptr );
        }

        else {
            mDataDev->setPrimarySelection( nullptr );
        }
    }

    switch ( type ) {
        case DesQ::Clipboard::Clipboard: {
            clipboard->promoteToTop( idx );
            WlrootsQt::MimeData mimeData = clipboard->entry( 0 );
            offerClipboard( mimeData );
            emit updateView( 0 );
            break;
        }

        case DesQ::Clipboard::Selection: {
            selection->promoteToTop( idx );
            WlrootsQt::MimeData mimeData = selection->entry( 0 );
            offerSelection( mimeData );
            emit updateView( 1 );
            break;
        }
    }
}


void DesQ::Clipboard::Manager::handleClipboardOffer( WlrootsQt::DataControlOffer *offer ) {
    /** This is our own offer. Nothing needs to be done about this. */
    if ( offer->offeredMimeTypes().contains( "application/x-desq-clipboard" ) ) {
        /**
         * Leave in peace
         * This is wayland providing our own data back to us,
         * because, we're subscribed to all it's events.
         */

        return;
    }

    /** Alien data offer. Gather it. */
    WlrootsQt::MimeData mimeData = gatherOfferedData( offer );

    /**
     * Empty clipboard prevention:
     * Copy the mimeData to lastClipboard only if it was not empty.
     * Or if it was not large data.
     */
    if ( mimeData.formats().count() ) {
        if ( mimeData.formats().contains( "application/x-large-data" ) ) {
            int  size = mimeData.data( "application/x-large-data" ).toULongLong();
            emit largeData( size );

            return;
        }

        lastClipboard = mimeData;
    }

    clipboard->addEntry( lastClipboard );
    emit updateView( 0 );

    offerClipboard( lastClipboard );
}


void DesQ::Clipboard::Manager::handleSelectionOffer( WlrootsQt::DataControlOffer *offer ) {
    /**
     * Ignore Clipboard
     * We will not re-offer the Primary Selection data.
     * We will ignore the data offer and return immediately.
     */
    if ( clipSett->value( "IgnoreSelections" ) ) {
        return;
    }

    /** This is our own offer. Nothing needs to be done about this. */
    if ( offer->offeredMimeTypes().contains( "application/x-desq-clipboard" ) ) {
        /**
         * Leave in peace
         * This is wayland providing our own data back to us,
         * because, we're subscribed to all it's events.
         */

        return;
    }

    /**
     * If we are restoring the session and (Primary)Selections were disabled in the last
     * Session, we will again Disable Selections.
     */
    if ( DesQ::Clipboard::disableSelections ) {
        /** Destroy the old data source */
        mDataDev->setPrimarySelection( nullptr );

        /** Return without gathering data or anything else */
        return;
    }

    /** Alien data offer. Gather it. */
    WlrootsQt::MimeData mimeData = gatherOfferedData( offer );

    /**
     * Empty clipboard prevention:
     * Copy the mimeData to lastSelection only if it was not empty.
     * Or if it was not large data.
     */
    if ( mimeData.formats().count() ) {
        if ( mimeData.formats().contains( "application/x-large-data" ) ) {
            int  size = mimeData.data( "application/x-large-data" ).toULongLong();
            emit largeData( size );

            return;
        }

        lastSelection = mimeData;
    }

    selection->addEntry( lastSelection );
    emit updateView( 1 );

    offerSelection( lastSelection );
}


void DesQ::Clipboard::Manager::offerClipboard( WlrootsQt::MimeData mimeData ) {
    /**
     * This is the offer part of the Manager. Once the data is gathered,
     * It will be sent here for offering it. If the mimeData collected is null,
     * we will offer the last known good data.
     *
     * Applicable to: Clipboard (Wlroots::Selection)
     *
     * Prevent Empty Clipboard:
     * Apps like firefox are sensitive and sometimes can cause clipboard problems.
     * Similarly, if an app providing the data closes/crashes, the user will be
     * left with an empty clipboard. To prevent such cases, we will re-offer the
     * data we've recieved.
     *
     * BUG: Currently, there is no way to disable this "feature"
     */

    if ( mimeData.formats().count() ) {
        if ( not (lastClipboard == mimeData) ) {
            lastClipboard = mimeData;
        }
    }

    /** Destroy the old data source */
    mDataDev->setSelection( nullptr );

    /** Create a new data source, and set the mimeData */
    clpSrc = mDataMgr->createDataSource();

    clpSrc->setSelectionData( lastClipboard );

    /** Offer the mimeTypes */
    clpSrc->offer( "application/x-desq-clipboard" );
    for ( QString fmt: mimeData.formats() ) {
        clpSrc->offer( fmt );
    }

    /** Set this source as the new selection to the data_device */
    mDataDev->setSelection( clpSrc );
}


void DesQ::Clipboard::Manager::offerSelection( WlrootsQt::MimeData mimeData ) {
    /**
     * This is the offer part of the Manager. Once the data is gathered,
     * It will be sent here for offering it. If the mimeData collected is null,
     * we will offer the last known good data.
     *
     * Applicable to: Selection (Wlroots::PrimarySelection)
     *
     * Prevent Empty Clipboard:
     * Apps like firefox are sensitive and sometimes can cause clipboard problems.
     * Similarly, if an app providing the data closes/crashes, the user will be
     * left with an empty clipboard. To prevent such cases, we will re-offer the
     * data we've recieved.
     *
     * BUG: Currently, there is no way to disable this "feature"
     */

    if ( mimeData.formats().count() ) {
        if ( not (lastSelection == mimeData) ) {
            lastSelection = mimeData;
        }
    }

    /** Destroy the old data source */
    mDataDev->setPrimarySelection( nullptr );

    /** Create a new data source, and set the mimeData */
    selSrc = mDataMgr->createDataSource();

    selSrc->setSelectionData( lastSelection );

    /** Offer the mimeTypes */
    selSrc->offer( "application/x-desq-clipboard" );
    for ( QString fmt: mimeData.formats() ) {
        selSrc->offer( fmt );
    }

    /** Set this source as the new selection to the data_device */
    mDataDev->setPrimarySelection( selSrc );
}


WlrootsQt::MimeData DesQ::Clipboard::Manager::gatherOfferedData( WlrootsQt::DataControlOffer *offer ) {
    /**
     * Since we provide the "Prevent Empty Clipboard feature", we have an associated problem.
     * Some apps like image-editors can overload the clipboard will large data.
     * For example:
     * Copying an image with gimp will result in a copy buffer of 100+ MiB data per format,
     * with multiple formats made available: image/bmp, image/jpeg, image/png, image/tiff,
     * image/x-MS-bmp, image/x-bmp. Out of these, image/bmp, image/x-MS-bmp and image/x-bmp
     * are the same format and can be very large (several tens of MiB to even 100+ MiB).
     *
     * These situations can cause memory problems. To prevent such issues, we're giving the
     * option to the user to limit the memory usage. In case the size of the data exceeds a
     * particular limit, then, mimeData will contain only one format `application/x-large-data`,
     * and the data will be the size of the data recieved.
     * When the handlers recieve this format, they will not try to store/re-offer the data.
     */

    WlrootsQt::MimeData mimeData;
    int                 dataSize = 0;

    if ( offer and offer->isValid() ) {
        QStringList mimeTypes( offer->offeredMimeTypes() );
        qDebug() << mimeTypes << "were offered";
        QByteArray  previous;
        for ( QString mt: mimeTypes ) {
            if ( not offer->isValid() ) {
                qDebug() << "Invalidated offer";
            }

            qDebug() << "Asking for" << mt;
            QByteArray data = offer->retrieveData( mt );

            /** We have some valid data */
            if ( data.count() ) {
                mimeData.setData( mt, data );
                previous = data;
                qDebug() << data;
            }

            /** We have previously retrieved data */
            else if ( previous.count() ) {
                mimeData.setData( mt, data );
                previous = data;
            }

            /** Count the size of the data in the mimeData object */
            dataSize += previous.count();
        }

        QObject::connect(
            offer, &WlrootsQt::DataControlOffer::invalidated, [ = ] () {
                qDebug() << "Disconnected";
                offer->disconnect();
            }
        );
    }

    if ( (bool)clipSett->value( "LimitSize" ) ) {
        int limit = (int)clipSett->value( "MaxDataSize" );

        if ( dataSize > limit ) {
            mimeData.clear();
            mimeData.setData( "application/x-large-data", QByteArray::number( dataSize ) );
        }
    }

    return mimeData;
}
